package pl.webnet.JaTrac;

import java.awt.*;
import java.util.*;

/**
 * This class describes a generic primitive - ancestor of all
 * object types that can be intersected by a ray.
 * @author: Miloslaw Smyk (thorgal@amiga.com.pl)
 * @version: 1.0
 */


abstract class Primitive
{
	Spectrum spectrum = null;
	Surface surface = null;
	Mapping mapping = null;
	String name = null;

/**
 * Check if the ray intersects this primitive and compute intersection points
 * as a value of <code>t</code> along the ray.
 * @param		ray the ray to test against
 * @return	the number of intersections
 */
	abstract public int intersect(Ray ray);

/**
 * Compute normal vector at given point on object's surface.
 * @param		p point (in world coordinates) on object's surface
 * @return	normal vector
 */
	abstract public Vec getNormal(Point p);
	
/**
 * Compute tangent vector at given point on object's surface.
 * Used for bump mapping.
 * @param		n surace's normal vector at this point. If null, will be computed.
 * @param		p point (in world coordinates) on object's surface
 * @return	normal vector
 */
	abstract public Vec getTangent(Point p, Vec normal);

	public Surface getSurface() { return surface; }
	
	public void setSurface(Surface surface) { this.surface = surface;	}

	public Spectrum getSpectrum() { return spectrum; }

	public void setSpectrum(Spectrum s)	{ spectrum = s;	}

	public void setName(String name) { this.name = name; }

	public void setParam(String param, String value)
	{
	}
}
