package pl.webnet.JaTrac;

import java.util.*;
import java.awt.*;

class RayTree
{
	Scene scene;
	int maxRecurseLevel;
	Texture texture2 = new GradientColorMap(Color.white, Color.magenta, Color.blue, Color.yellow);
	//Texture texture2 = new ImageColorMap("/home/thorgal/code/jatrac/earthmap.jpg");
	//Texture texture = new ImageBumpMap("/home/thorgal/code/jatrac/earthmap.jpg");
	Texture texture = new ImageBumpMap("/home/thorgal/code/jatrac/amiga.jpg");

	public RayTree(Scene s, int maxLevel)
	{
		scene = s;
		maxRecurseLevel = maxLevel;
	}

	public Ray trace(int level, double weight, Ray ray)
	{
		Primitive prim;
		Point hitPoint;
		Vec normalVec;
		Intersection intersection;
		
		if(intersect(ray) != 0)
		{
			intersection = (Intersection)ray.intersections.first();

			hitPoint = ray.rayPoint(intersection.t);
			normalVec = intersection.primitive.getNormal(hitPoint);

			if(ray.dot(normalVec) > 0.)
				normalVec.negate();

			ray.spectrum = shade(level, weight, hitPoint, normalVec, ray, intersection);
		}
		else
			if(level == 0)
				ray.spectrum = scene.spectrum.create(88, 35, 149);
			else
				ray.spectrum = scene.spectrum.create(128, 0, 0);


		return ray;
	}

	public Spectrum shade(int level, double weight, Point hitPoint, Vec normalVec, Ray ray, Intersection intersection)
	{
		Primitive prim = intersection.primitive;
		Spectrum diffPrimSpectrum, specPrimSpectrum;
		Spectrum primSpectrum = prim.getSpectrum();
		Spectrum reflSpectrum;
		Spectrum diffSpectrum = scene.spectrum.create(Spectrum.black);
		Spectrum specSpectrum = scene.spectrum.create(Spectrum.black);

		double LRdot;
		double lightAngle;
		Double shadowCache;
		float[] cols;
		Surface surface = prim.getSurface();

		if(prim.mapping != null)
		{
			Point uv = prim.mapping.getUV(hitPoint, normalVec);
			Vec tangent = prim.getTangent(hitPoint, normalVec);

			if(uv != null)
			{
				normalVec = texture.updateNormal(normalVec, tangent, uv);
				primSpectrum = texture2.updateSpectrum(primSpectrum, uv);
			}
		}
		
		diffPrimSpectrum = primSpectrum.create(primSpectrum);
		diffPrimSpectrum.scale(surface.diffuseCoeff);

		specPrimSpectrum = primSpectrum.create(primSpectrum);
		specPrimSpectrum.scale(surface.specularCoeff);

		// create unit specular vector
		Ray specularRay = generateSpecularRay(hitPoint, normalVec, ray);
		specularRay.unit();

		// iterate over all light sources gathering diffuse light and highlights
		for(Iterator i = scene.lights.iterator(); i.hasNext(); )
		{
			shadowCache = null;
			Light light = (Light)i.next();

			Vec lightVec = new Vec(light.getOrigin(), hitPoint);
			double lightVecLen = lightVec.length();

			lightVec.unit();

			if((lightAngle = normalVec.dot(lightVec)) > 0.)
			{
				if((shadowCache = light.shadow(this, hitPoint, lightVec, lightVecLen)).doubleValue() != 0.)
					// multiply and accumulate
					diffSpectrum.MAC(lightAngle * shadowCache.doubleValue(), light.getSpectrum());
			}

			LRdot = specularRay.dot(lightVec);

			if(LRdot > 0.)
			{
				//if(shadowCache == null && (shadowCache = shadow(new Ray( hitPoint, lightVec), lightVecLen)) == null)
				//if(shadowCache == null && (shadowCache = light.shadow(this, hitPoint, lightVec, lightVecLen)).doubleValue() != 0.)
				shadowCache = light.shadow(this, hitPoint, lightVec, lightVecLen);
				{
					double specular = shadowCache.doubleValue() * Math.pow(specularRay.dot(lightVec), surface.specularSizeCoeff);
					specSpectrum.brighten(specular);
				}
			}
		}

		// calculate the actual fraction of diffuse and specular spectrum reflected
		// from the surface
		diffSpectrum.interact(diffPrimSpectrum);
		specSpectrum.interact(specPrimSpectrum);

		if(level + 1 <= maxRecurseLevel)
		{
			if(surface.reflectionCoeff * weight > .05)
			{
				ray.reflectedRay = trace(level + 1, surface.reflectionCoeff * weight, specularRay);

				reflSpectrum = ray.reflectedRay.spectrum.create(ray.reflectedRay.spectrum);
				reflSpectrum.interact(surface.reflectionCoeff, primSpectrum);
//				diffSpectrum.scale(1 - surface.reflectionCoeff);
				diffSpectrum.add(reflSpectrum);
			}
/*

			if(surface.transmissionCoeff != 0.)
			{
				Ray tr = generateTransmittedRay(hitPoint, normalVec, ray);

				if(tr == null)
					tr = specularRay;
				ray.transmitted = trace(level + 1, weight, tr);

				r = rs + rd * (1 - surface.transmissionCoeff) + cols[0] * surface.transmissionCoeff * (ray.transmitted.color.getRed() / 255.) * weight;
				g = gs + gd * (1 - surface.transmissionCoeff) + cols[1] * surface.transmissionCoeff * (ray.transmitted.color.getGreen() / 255.) * weight;
				b = bs + bd * (1 - surface.transmissionCoeff) + cols[2] * surface.transmissionCoeff * (ray.transmitted.color.getBlue() / 255.) * weight;

				}
*/
		}
			
		diffSpectrum.add(specSpectrum);


		return diffSpectrum;
	}

	public Ray generateSpecularRay(Point hitPoint, Vec normalVec, Vec rayDir)
	{
		return new Ray(hitPoint, normalVec.addScalarMultiple(-2. * normalVec.dot(rayDir), rayDir));
	}

	public Ray generateTransmittedRay(Point hitPoint, Vec normalVec, Ray incidentRay)
	{
		double ri1 = 1.0, ri2 = 1.0;
	/*	
		Surface m[] = new Surface[2];

		for(Iterator i =  (Intersection)incidentRay.intersections.iterator(); i.hasNext(); )
		{
			Intersection is = (Intersection)i.next();
			m[idx++] = is.surface;

			if(idx == 2)
				break;
		}

		if(m[0]
		*/

		Surface s = ((Intersection)incidentRay.intersections.first()).surface;
		
		if(s != null)
			ri1 = s.refractionIndex;

		double eta = ri1 / ri2;
		double c1 = -incidentRay.dot(normalVec);
		double cs2 = 1. - eta * eta * (1. - c1 * c1);

		if(cs2 < 0.)
			return null;

		return (new Ray(hitPoint, incidentRay.linearCombination(eta, eta * c1 - Math.sqrt(cs2), normalVec)));
	}

	public int intersect(Ray ray)
	{
		int hits = 0;
		
		for(Iterator i = scene.primitives.iterator(); i.hasNext(); )
		{
			Primitive p = (Primitive)i.next();
			
			hits += p.intersect(ray);
		}

		return hits;
	}
	
}
