package pl.webnet.JaTrac;

import java.awt.*;

class Triangle extends Primitive
{
	static private final int AXIS_X = 1;
	static private final int AXIS_Y = 2;
	static private final int AXIS_Z = 3;

	Point x1, x2, x3;
	Point P1, P2, P3;
	Vec normal;
	Vec tangent;
	double D;
	double area;
	int dominant;

	public Triangle(Point x1, Point x2, Point x3)
	{
		this.x1 = new Point(x1);
		this.x2 = new Point(x2);
		this.x3 = new Point(x3);

		Vec edge1 = new Vec(x1.x - x2.x, x1.y - x2.y, x1.z - x2.z);
		Vec edge2 = new Vec(x3.x - x2.x, x3.y - x2.y, x3.z - x2.z);

		normal = edge1.cross(edge2);
		normal.unit();

		tangent = edge1;
		tangent.unit();

		D = - (normal.x * x2.x + normal.y * x2.y + normal.z * x2.z);

		double nx = Math.abs(normal.x);
		double ny = Math.abs(normal.y);
		double nz = Math.abs(normal.z);

		// select plane parallel to one of the axes and map the triangle to it
		if(nx >= ny && nx >= nz)
		{
			P1 = new Point(x1.y, x1.z, 0.);
			P2 = new Point(x2.y, x2.z, 0.);
			P3 = new Point(x3.y, x3.z, 0.);
			dominant = AXIS_X;
		}
		else
			if(ny >= nx && ny >= nz)
			{
				P1 = new Point(x1.x, x1.z, 0.);
				P2 = new Point(x2.x, x2.z, 0.);
				P3 = new Point(x3.x, x3.z, 0.);
				dominant = AXIS_Y;
			}
			else
				if(nz >= nx && nz >= ny)
				{
					P1 = new Point(x1.x, x1.y, 0.);
					P2 = new Point(x2.x, x2.y, 0.);
					P3 = new Point(x3.x, x3.y, 0.);
					dominant = AXIS_Z;
				}
		

		// compute area of the 2D triangle created above
		area = Math.abs(P1.x * P2.y - P1.y * P2.x + P2.x * P3.y - P3.x * P2.y + P3.x * P1.y - P1.x * P3.y);

		System.out.println("Triangle added: " + this);
	}


	public int intersect(Ray ray)
	{
		double vd = normal.dot(ray);
		double x = 0, y = 0;

		// check if ray is not parallel to the triangle's plane
		if(vd == 0.)
			return 0;

		Vec rayOrigin = new Vec(ray.origin);
//		rayOrigin.unit();

		double v0 = - (normal.dot(rayOrigin) + D);
		double t = v0 / vd;

		// check if the intersection is not behind ray's origin
		if(t < Ray.EPSILON)
			return 0;

		// find coordinates of intersection point in 2D triangle's space
		switch(dominant)
		{
			case AXIS_X:
				x = ray.y * t + ray.origin.y;
				y = ray.z * t + ray.origin.z;
				break;

			case AXIS_Y:
				x = ray.x * t + ray.origin.x;
				y = ray.z * t + ray.origin.z;
				break;

			case AXIS_Z:
				x = ray.x * t + ray.origin.x;
				y = ray.y * t + ray.origin.y;
				break;
		}

		double tempArea;

		if(area >= (tempArea = Math.abs(x * P2.y - y * P2.x + P2.x * P3.y - P3.x * P2.y + P3.x * y - x * P3.y)) &&
			area >= (tempArea += Math.abs(P1.x * y - P1.y * x + x * P3.y - P3.x * y + P3.x * P1.y - P1.x * P3.y)) &&
			Math.abs(area - (tempArea += Math.abs(P1.x * P2.y - P1.y * P2.x + P2.x * y - x * P2.y + x * P1.y - P1.x * y))) <= Ray.EPSILON)


//		if(Math.abs(area - (tempArea1 + tempArea2 + tempArea3)) <= Ray.EPSILON)
		{
		//System.out.println("as: " + area + ", " + tempArea1 + ", " + tempArea2 + ", " + tempArea3 + ". x: " + x + ", y: " + y + ". t=" + t);
		//System.out.println("Ray: " + ray.x + ", " + ray.y + ", " + ray.z);
			ray.intersections.add(new Intersection(t, this, Intersection.ENTERING, null));
			return 1;
		}

		return 0;
	}

	public Vec getNormal(Point p)
	{
		return new Vec(normal);		
	}

	public Vec getTangent(Point p, Vec n)
	{
		return new Vec(tangent);
	}

	public String toString()
	{
		return(getClass().getName() + '@' + Integer.toHexString(hashCode()) + ": " + x1 + ", " + x2 + ", " + x3 + ". Normal: " + normal + ". D: " + D + ". Dominant: " + dominant);
	}
}
