package pl.webnet.JaTrac;

import java.lang.*;

/**
 * This class handles 3D vectors.
 * @author: Miloslaw Smyk (thorgal@amiga.com.pl)
 * @version: 1.0
 */

class Vec
{
	protected double x, y, z;
	protected double length;
	boolean normalized = false;
	boolean measured = false;

	public Vec()
	{
	}
	
	public Vec(Vec a)
	{
		copy(a);		
	}

	public Vec(double x, double y, double z)
	{
		this.x = x;
		this.y = y;
		this.z = z;
	}

	public Vec(double[] abc)
	{
		x = abc[0];
		y = abc[1];
		z = abc[2];
	}

	public Vec(double x, double y, double z, boolean normalized)
	{
		this.x = x;
		this.y = y;
		this.z = z;
		this.normalized = normalized;
	}

	public Vec(Point e, Point s)
	{
		x = e.x - s.x;
		y = e.y - s.y;
		z = e.z - s.z;
	}

	public Vec(Point p)
	{
		x = p.x;
		y = p.y;
		z = p.z;
	}
	
	public void setXYZ(double x, double y, double z)
	{
		this.x = x;
		this.y = y;
		this.z = z;

		normalized = measured = false;
	}

	public void setX(double x)
	{
		this.x = x;
		normalized = measured = false;
	}
	
	public void setY(double y)
	{
		this.y = y;
		normalized = measured = false;
	}
	
	public void setZ(double z)
	{
		this.z = z;
		normalized = measured = false;
	}

	public double length()
	{
		if(measured != true)
		{
			length = StrictMath.sqrt(x * x + y * y + z * z);
			measured = true;
		}

		return length;
	}

	public double dot(Vec a)
	{
		Vec b = this;

		double dot = a.x * b.x + a.y * b.y + a.z * b.z;

		return dot;
	}

	public Vec cross(Vec b)
	{
		return new Vec(y * b.z - z * b.y, z * b.x - x * b.z, x * b.y - y * b.x);
	}

	public void mul(double t)
	{
		x *= t;
		y *= t;
		z *= t;
	}
	
	public double unit()
	{
		double len = length();

		if(normalized == false && len != 0.0)
		{
			x /= len;
			y /= len;
			z /= len;
			normalized = true;
			measured = false;
			length();
		}

		return len;
	}

	public void negate()
	{
		x = -x;
		y = -y;
		z = -z;
	}

	public Vec addScalarMultiple(double sm, Vec b)
	{
		return new Vec(sm * x + b.x, sm * y + b.y, sm * z + b.z);
	}

	public Vec addVectorMultiple(Vec a, double vm)
	{
		return new Vec(x + a.x * vm, y + a.y * vm, z + a.z * vm);
	}

	public Vec linearCombination(double sa, double sb, Vec b)
	{
		return new Vec(x * sa + b.x * sb, y * sa + b.y * sb, z * sa + b.z * sb);
	}

	public void copy(Vec a)
	{
		setXYZ(a.x, a.y, a.z);
		measured = a.measured;

		if(measured == true)
			length = a.length();

		normalized = a.normalized;
	}
	
	public String toString()
	{
		return(getClass().getName() + '@' + Integer.toHexString(hashCode()) + ": " + String.valueOf(x) + ", " + String.valueOf(y) + ", " + String.valueOf(z));
	}
}

